//*****************************************************************************
//  Copyright (c) 2021 Glenfly Tech Co., Ltd.. 
//  All Rights Reserved.
//
//  This is UNPUBLISHED PROPRIETARY SOURCE CODE of Glenfly Tech Co., Ltd..;
//  the contents of this file may not be disclosed to third parties, copied or
//  duplicated in any form, in whole or in part, without the prior written
//  permission of Glenfly Tech Co., Ltd..
//
//  The copyright of the source code is protected by the copyright laws of the People's
//  Republic of China and the related laws promulgated by the People's Republic of China
//  and the international covenant(s) ratified by the People's Republic of China. 
//*****************************************************************************

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include "gf_chip_id.h"
#include "gf_hwinfo.h"
#define DEBUSINFO "/sys/kernel/debug/dri/0/gf/info"



 int  gf_get_struct_size(int fd)
 {
    if (fd<0)
    {
        return 0;
    }
    else
    {
        return sizeof(gf_dev_t);
    }
    
 }

 static  DWORD SEND_VCP_QUERY_E3K(DWORD BlockId, DWORD Offset, DWORD DumpSize, large_inter  address, DWORD* pCmd)
{
   
    Vcp_Decouple_Opcodes_cmd         cmd_CSP_Vcp = {0};
    DWORD* pCmd0 = pCmd;
    cmd_CSP_Vcp.cmd_Vcp_Query_Dump.Major_Opcode = VCP_OPCODE_Vcp_Query_Dump;
    cmd_CSP_Vcp.cmd_Vcp_Query_Dump.Block_Id     = BlockId;
    cmd_CSP_Vcp.cmd_Vcp_Query_Dump.Dwc          = DumpSize;
    cmd_CSP_Vcp.cmd_Vcp_Query_Dump.Reg_Offset   = Offset;

    *pCmd++ = cmd_CSP_Vcp.uint;
    *pCmd++ = address.low32;
    *pCmd++ = address.high32;

    return (DWORD)(pCmd - pCmd0);
}

static DWORD  SEND_INTERNAL_FENCE(DWORD Block, HWM_SYNC_SLOT_E3K Slot, DWORD Value,DWORD* pCmd)
{
    DWORD * pCmd0 = pCmd;
    Vcp_Decouple_Opcodes_cmd     SendFenceCmd={0};
    Cmd_Vcp_Fence_Internal_Dword1     DW1={0};

    SendFenceCmd.cmd_Vcp_Fence.Major_Opcode        = VCP_OPCODE_Vcp_Fence;//0x8
    SendFenceCmd.cmd_Vcp_Fence.Block_Id            = Block;
    SendFenceCmd.cmd_Vcp_Fence.Slot_Id             = Slot;
    //SendFenceCmd.cmd_Vcp_Fence.Update_Value        = Value;
    SendFenceCmd.cmd_Vcp_Fence.Fence_Type          = VCP_FENCE_FENCE_TYPE_INTERNAL;    
    SendFenceCmd.cmd_Vcp_Fence.Fence_Update_Mode   = VCP_FENCE_FENCE_UPDATE_MODE_COPY;//0x0
    //SendFenceCmd.cmd_Vcp_Fence.Fence_Update_Timing = VCP_INTERNAL_FENCE_FENCE_UPDATE_TIMING_UPDATE_AFTER_WRITE;//0x1
    SendFenceCmd.cmd_Vcp_Fence.Dwc = 1;

    DW1.Update_Value = Value;

    *pCmd++ = (DWORD)SendFenceCmd.uint;
    *pCmd++ = *(DWORD*)(&DW1);

    

    return (DWORD)(pCmd-pCmd0); 
}

static DWORD  SEND_INTERNAL_WAIT(DWORD Block, HWM_SYNC_SLOT_E3K Slot, DWORD Value,DWORD* pCmd)
{
    DWORD * pCmd0 = pCmd;
    Vcp_Decouple_Opcodes_cmd     WaitCmd={0};

    WaitCmd.cmd_Vcp_Internal_Wait.Major_Opcode = VCP_OPCODE_Vcp_Internal_Wait;
    WaitCmd.cmd_Vcp_Internal_Wait.Slot_Id      = Slot;
    WaitCmd.cmd_Vcp_Internal_Wait.Ref_Value    = Value;
    WaitCmd.cmd_Vcp_Internal_Wait.Fe_Be_Flag   = Block;
    WaitCmd.cmd_Vcp_Internal_Wait.Wait_Mode    = VCP_INTERNAL_WAIT_WAIT_MODE_NORMAL_WAIT;
    WaitCmd.cmd_Vcp_Internal_Wait.Station_Id   = VCP_INTERNAL_WAIT_STATION_ID_MAIN_PARSER;
    WaitCmd.cmd_Vcp_Internal_Wait.Method       = WAIT_METHOD_BIGEQUAL;


    *pCmd++ = (DWORD)WaitCmd.uint;

    return (DWORD)(pCmd-pCmd0); 
}

static DWORD  SET_BCI_REG(DWORD uBlockId, DWORD uStartOffset, DWORD uSize, DWORD *pData,DWORD* pCmd)
{
    DWORD * pCmd0 = pCmd;
    Vcp_Decouple_Opcodes_cmd         cmd_CSP_Vcp = {0};
    memset(&cmd_CSP_Vcp, 0, sizeof(cmd_CSP_Vcp));

    //=========================Set Register====================================
    cmd_CSP_Vcp.cmd_Vcp_Set_Register.Major_Opcode         = VCP_OPCODE_Vcp_Set_Register;
    cmd_CSP_Vcp.cmd_Vcp_Set_Register.Block_Id             = uBlockId;
    cmd_CSP_Vcp.cmd_Vcp_Set_Register.Start_Offset         = uStartOffset;   
    cmd_CSP_Vcp.cmd_Vcp_Set_Register.Dwc                  = uSize;   // Set Register Following DW Number
    *pCmd++ = cmd_CSP_Vcp.uint;                               // Set Register CMD header

    memcpy(pCmd, pData, uSize * 4);                           //Set Register CMD data
    pCmd += uSize;
    return (DWORD)(pCmd-pCmd0);
}

int gf_read_reg(int fd,unsigned int offset,unsigned int len,void *buf)
{
    gf_query_info_t info;
    if(buf)
    {
        info.argu=offset;
        info.type=GF_QUERY_REGISTER_U32;
        info.buf_len=len;
        info.buf=buf;

    }
    return gfQueryInfo(fd,&info);

}
int gf_set_reg(int fd, unsigned int offset,unsigned int value)
{
    gf_query_info_t info;
   
    info.argu=offset;
    info.type=GF_SET_REGISTER_U32;
    info.value=value;
    //printf("value=%x\n",value);

    return gfQueryInfo(fd,&info);
}

int gf_init_hwinfo(int fd, void *pdev)
{
    gf_create_device_t createDevice = { 0 };
    gf_bo_alloc_arg_t boAllocArg = { 0 };
    gf_dev_t *dev=(gf_dev_t*)pdev;
    gf_query_info_t query_info = {0};

    Reg_Csp_Misc_Control     reg_Csp_Misc_Control      ={0};
    int ret=0;

    memset(dev,0,sizeof(gf_dev_t));  
    if(fd<0)
    {
           
        return -1;

    }

    query_info.type = GF_QUERY_CHIP_ID;
    gfQueryInfo(fd, &query_info);
    dev->ChipID = query_info.value;

    switch (dev->ChipID)
    {
    case CHIP_ARISE1020:  
    case CHIP_ARISE1010:  
        dev->RefCounter_Offset = MMIO_Csp_Ref_Total_Gpu_Timestamp_Offset;
        break;
//  case CHIP_ELT3K:
    case CHIP_ARISE:
    case CHIP_ARISE1040:
//  case CHIP_ARISE1020S:
        dev->RefCounter_Offset = MMIO_Csp_Ms_Total_Gpu_Timestamp_Offset;
        break;    
    default:
        printf("not suport chip id \n");
        break;
    }


    dev->gfFd=fd;
    ret=gfCreateDevice(dev->gfFd, &createDevice);
    dev->gfDevice = createDevice.device;
    dev->bufmgr = gf_bufmgr_init(dev->gfFd);
    dev->renderContext = gf_create_context(dev->gfFd, dev->gfDevice, RB_INDEX_VCP0);
    if (dev->bo==0) 
    {
        boAllocArg.width = 4096;
        boAllocArg.height = 1;
        boAllocArg.tiled = 0;
        boAllocArg.format = GF_FORMAT_UINT32;
        boAllocArg.usage_mask = GF_USAGE_TEMP_BUFFER;
        boAllocArg.access_hint = GF_ACCESS_GPU_ALMOST;
        boAllocArg.unpagable = 0;
        boAllocArg.tiled= 0;
        dev->bo = gf_bo_alloc(dev->bufmgr, &boAllocArg);          
    
    }

    dev->counter.hCounterAlloc=dev->bo->allocation;
    dev->counter.gpu_virt_addr=dev->bo->gpu_virt_addr;
    dev->counter.mode=1;
    dev->counter.CounterSize=HWINFO_COUTER_SIZE;


    gf_read_reg(fd,MMIO_Csp_Misc_Control_Offset,1,&reg_Csp_Misc_Control);
    reg_Csp_Misc_Control.reg.Ila_Mode=1 & 0xf;
    gf_set_reg(fd,MMIO_Csp_Misc_Control_Offset,reg_Csp_Misc_Control.uint);

    gf_get_hwinfo (fd,pdev, NULL);

    return ret;
}


int  gf_get_hwinfo (int fd,void *pdev, gf_meminfo *pmeminfo)
{
    
    DWORD* pCmdBuf;
    DWORD  CmdSize          =0;
    DWORD  AllocOffset      =0;
  
    gf_dev_t *dev          =(gf_dev_t*)pdev;
    
    counter_t *pcounter    =&(dev->counter);

    DWORD  allocation       =dev->bo->allocation;
    DWORD  CounterSize      =pcounter->CounterSize;
    DWORD  ILACounterMode   =pcounter->mode;
    DWORD* pData;

    Reg_Csp_Misc_Control     reg_Csp_Misc_Control      ={0};
    Reg_Set_Reg_1            Reg_1                     ={0};
    DWORD uRegData[3]={0};

    DWORD  counterval[HWINFO_COUTER_SIZE]={0};
    DWORD  counter[HWINFO_COUTER_SIZE]={0};
    unsigned long long      timestamp=0;
    unsigned int            eclk=dev->eclk;
    static DWORD fenceid=0;
    int offset=0;

 
    map_args_t              arg                  = { 0};
    int i=0;

    if(eclk==0)
    {
        eclk=300;
    }
     
    CmdSize += 16+16+8;
   
    pCmdBuf= gf_context_request_space(dev->renderContext, CmdSize ,GF_CMD_HINT_3DBLT);

    pcounter->virt_addr=NULL;
    

    gf_bo_wait_idle(dev->bo);

    gf_bo_map(dev->bo, &arg);

    pcounter->virt_addr = arg.virt_addr;
    memset(pcounter->virt_addr,0,4096*sizeof(DWORD));
    gf_bo_unmap(dev->bo);

    Reg_1.reg.Mem_Check_Enable=0;

    pCmdBuf +=SET_BCI_REG(VCP_QUERY_DUMP_BLOCK_ID_VCP_FE, Reg_1_Offset, 1, &Reg_1.uint, pCmdBuf);
    pCmdBuf +=SET_BCI_REG(VCP_QUERY_DUMP_BLOCK_ID_VCP_BE, Reg_1_Offset, 1, &Reg_1.uint, pCmdBuf);

   
    fenceid++;
    pCmdBuf +=SEND_INTERNAL_FENCE(VCP_QUERY_DUMP_BLOCK_ID_VCP_FE,HWM_SYNC_VCP0_FE_SLOT,fenceid,pCmdBuf);
    pCmdBuf +=SEND_INTERNAL_FENCE(VCP_QUERY_DUMP_BLOCK_ID_VCP_BE,HWM_SYNC_VCP0_BE_SLOT,fenceid,pCmdBuf);

    pCmdBuf +=SET_BCI_REG(VCP_QUERY_DUMP_BLOCK_ID_VCP_FE, Reg_1_Offset, 1, &Reg_1.uint, pCmdBuf);
    pCmdBuf +=SET_BCI_REG(VCP_QUERY_DUMP_BLOCK_ID_VCP_BE, Reg_1_Offset, 1, &Reg_1.uint, pCmdBuf);

    fenceid++;
    pCmdBuf +=SEND_INTERNAL_FENCE(VCP_QUERY_DUMP_BLOCK_ID_VCP_FE,HWM_SYNC_VCP0_FE_SLOT,fenceid,pCmdBuf);
    pCmdBuf +=SEND_INTERNAL_FENCE(VCP_QUERY_DUMP_BLOCK_ID_VCP_BE,HWM_SYNC_VCP0_BE_SLOT,fenceid,pCmdBuf);


    fenceid++;
    AllocOffset=0;

    gf_context_add_allocation(dev->renderContext,allocation,pCmdBuf+1,AllocOffset, 0, 1, 1, MM_DRIVERID_VIDEO_RESERVEDMEM);
     
    pCmdBuf += SEND_VCP_QUERY_E3K(VCP_QUERY_DUMP_BLOCK_ID_VCP_FE, Reg_Fe_Framelvl_Cmp_Memcount_Offset, 
                                      S3VD_FE_CMP_COUNTER_SIZE, (large_inter)(dev->bo->gpu_virt_addr+AllocOffset), pCmdBuf);
                                     
                                
    pCmdBuf +=SEND_INTERNAL_FENCE(VCP_QUERY_DUMP_BLOCK_ID_VCP_FE,HWM_SYNC_VCP0_FE_SLOT,fenceid,pCmdBuf);

    AllocOffset+=S3VD_FE_CMP_COUNTER_SIZE*sizeof(DWORD);

    gf_context_add_allocation(dev->renderContext,allocation,pCmdBuf+1,AllocOffset, 0, 1, 1, MM_DRIVERID_VIDEO_RESERVEDMEM);

    pCmdBuf += SEND_VCP_QUERY_E3K(VCP_QUERY_DUMP_BLOCK_ID_VCP_BE, Reg_Be_Framelvl_Cmp_Memcount_Offset,
    S3VD_BE_CMP_COUNTER_SIZE, (large_inter)(dev->bo->gpu_virt_addr+AllocOffset), pCmdBuf);
    

 
    pCmdBuf +=SEND_INTERNAL_FENCE(VCP_QUERY_DUMP_BLOCK_ID_VCP_BE,HWM_SYNC_VCP0_BE_SLOT,fenceid,pCmdBuf);


    fenceid++;
    pCmdBuf +=SEND_INTERNAL_FENCE(VCP_QUERY_DUMP_BLOCK_ID_VCP_FE,HWM_SYNC_VCP0_FE_SLOT,fenceid,pCmdBuf);
    pCmdBuf +=SEND_INTERNAL_FENCE(VCP_QUERY_DUMP_BLOCK_ID_VCP_BE,HWM_SYNC_VCP0_BE_SLOT,fenceid,pCmdBuf);

    Reg_1.reg.Mem_Check_Enable=1;
    pCmdBuf +=SET_BCI_REG(VCP_QUERY_DUMP_BLOCK_ID_VCP_FE, Reg_1_Offset, 1, &Reg_1.uint, pCmdBuf);
    pCmdBuf +=SET_BCI_REG(VCP_QUERY_DUMP_BLOCK_ID_VCP_BE, Reg_1_Offset, 1, &Reg_1.uint, pCmdBuf);

   
    gf_context_release_space(dev->renderContext, pCmdBuf);

    gf_context_flush(dev->renderContext);

    gf_bo_wait_idle(dev->bo);

    gf_read_reg(fd, dev->RefCounter_Offset, MS_COUNTER_SIZE, &counterval[0]);

    gf_read_reg(fd, MMIO_Ila_Counters_Offset, ILA_COUNTER_SZIE, &counterval[MS_COUNTER_SIZE]);

    arg.virt_addr=NULL;
    gf_bo_map(dev->bo, &arg);

    pcounter->virt_addr = arg.virt_addr;

    if(pcounter->virt_addr)
    {
        pData=(DWORD *)pcounter->virt_addr;

        offset=MS_COUNTER_SIZE+ILA_COUNTER_SZIE;

        for(i=0;i<S3VD_COUNTER_SIZE;i++)
        {
            counterval[i+offset]=*(pData+i);
        }

    }

    if(pmeminfo)
    {    

        timestamp=*(unsigned long long*)counterval - *(unsigned long long*)(pcounter->counterval);

        //printf("timestamp=%llu\n",timestamp);

        for(i=MS_COUNTER_SIZE; i<HWINFO_COUTER_SIZE; i++)
        {
            counter[i]=(counterval[i]>=pcounter->counterval[i]) ? (counterval[i]-pcounter->counterval[i]) : (0xffffffff-pcounter->counterval[i]+counterval[i]);
        }

        pmeminfo->diu_rd_db=counter[MS_COUNTER_SIZE+25]/8*512llu/(timestamp/27);

        pmeminfo->diu_wr_db=0;

        pmeminfo->local_rd_db=counter[MS_COUNTER_SIZE+13]/8*256llu/(timestamp/27);
        pmeminfo->local_wr_db=counter[MS_COUNTER_SIZE+14]/8*256llu/(timestamp/27);      


        pmeminfo->vcp_rd_db=(counter[MS_COUNTER_SIZE+ILA_COUNTER_SZIE]/8*256llu/(timestamp/27))+(counter[MS_COUNTER_SIZE+ILA_COUNTER_SZIE+2]/8*256llu/(timestamp/27));

        pmeminfo->vcp_wr_db=(counter[MS_COUNTER_SIZE+ILA_COUNTER_SZIE+1]/8*256llu/(timestamp/27))+(counter[MS_COUNTER_SIZE+ILA_COUNTER_SZIE+3]/8*256llu/(timestamp/27));
        
        pmeminfo->vpp_rd_db=counter[MS_COUNTER_SIZE+28]/8*512llu/(timestamp/27);
        pmeminfo->vpp_wr_db=counter[MS_COUNTER_SIZE+29]/8*256llu/(timestamp/27);


        pmeminfo->gfx_rd_db=pmeminfo->local_rd_db + counter[MS_COUNTER_SIZE+36]/8*256llu/(timestamp/27) + counter[MS_COUNTER_SIZE+38]/8*256llu/(timestamp/27)
                            -pmeminfo->diu_rd_db - pmeminfo->vcp_rd_db - pmeminfo->vpp_rd_db;
	
        pmeminfo->gfx_wr_db=pmeminfo->local_wr_db + counter[MS_COUNTER_SIZE+37]/8*256llu/(timestamp/27) + counter[MS_COUNTER_SIZE+39]/8*256llu/(timestamp/27)
	                        - pmeminfo->diu_wr_db -pmeminfo->vcp_wr_db - pmeminfo->vpp_wr_db;
        
	
        pmeminfo->pcie_rd_db= counter[MS_COUNTER_SIZE+36]/8*256llu/(timestamp/27) + counter[MS_COUNTER_SIZE+38]/8*256llu/(timestamp/27);
	
        pmeminfo->pcie_wr_db= counter[MS_COUNTER_SIZE+37]/8*256llu/(timestamp/27) + counter[MS_COUNTER_SIZE+39]/8*256llu/(timestamp/27);

        
 /*       
        printf(" snoop_rd  = %lluMB/s\n", counter[MS_COUNTER_SIZE+36]/8*256llu/(timestamp/27));
        printf(" snoop_wr  = %lluMB/s\n", counter[MS_COUNTER_SIZE+37]/8*256llu/(timestamp/27));
        printf(" bius_rd   = %lluMB/s\n", counter[MS_COUNTER_SIZE+38]/8*256llu/(timestamp/27));
        printf(" bius_wr   = %lluMB/s\n", counter[MS_COUNTER_SIZE+39]/8*256llu/(timestamp/27));
 */   
    }

    for(i=0; i<HWINFO_COUTER_SIZE; i++)
    {     
        pcounter->counterval[i]=counterval[i];
    }
    pcounter->gputimestamp=timestamp;   

    gf_bo_unmap(dev->bo);

    return 0;

}



 void gf_deinit_hwinfo(int fd, void *pdev)
 {
    gf_dev_t *dev          =(gf_dev_t*)pdev;

    gf_bo_unreference(dev->bo);
    gf_destroy_context(dev->renderContext);
    gfDestroyDevice(fd,dev->gfDevice);
 }

